/**********************************************************************
 * Copyright (c) 2006 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * IBM - Initial API and implementation
 **********************************************************************/
package org.eclipse.cosmos.rm.validation.internal.databuilders;

import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;

import org.eclipse.cosmos.rm.validation.internal.SMLActivator;
import org.eclipse.cosmos.rm.validation.internal.common.ISMLConstants;
import org.eclipse.cosmos.rm.validation.internal.common.IValidationConstants;
import org.eclipse.cosmos.rm.validation.internal.smlvalidators.ValidationMessages;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

/**
 * Used to build an inheritance map that will indicate the relationships between
 * the reference types.
 * 
 * @author sleeloy
 * @author Ali Mehregani
 */
public class TypeInheritanceDataBuilderImpl extends AbstractDataBuilder implements IDataBuilder 
{
	
	/**
	 * The static ID of this builder
	 */
	public static final String ID = SMLActivator.PLUGIN_ID + ".TypeInheritanceDataBuilderImpl";

	/**
	 * The inheritance map
	 * 
	 * KEY = An object of type {@link String} indicating a reference
	 * VALUE = An object of type {@link String} indicating the immediate parent of KEY 
	 */
	private Map inheritanceMap = new HashMap();

	/**
	 * the current complex type that is being parsed;
	 */
	private String currentComplexType;

	/**
	 * Stores the attributes of a complex type
	 * KEY = Complex type name
	 * VALUE = The attributes of the complex type
	 */
	private Map attributes;
	
	/**
	 * Constructor
	 */
	public TypeInheritanceDataBuilderImpl()
	{
		attributes = new Hashtable();
	}
	
	/**
	 * The data structure is a {@link Map}, where 
	 * 
	 * KEY = An object of type {@link String} indicating a reference
	 * VALUE = An object of type {@link String} indicating the immediate parent of KEY
	 *  
	 * @see org.eclipse.cosmos.rm.validation.internal.databuilders.IDataBuilder#getDataStructure()
	 */
	public Object getDataStructure()
	{
		return inheritanceMap;
	}

	
	/**
	 * @see org.xml.sax.helpers.DefaultHandler#startElement(java.lang.String, java.lang.String, java.lang.String, org.xml.sax.Attributes)
	 */
	public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
	{
		super.startElement(uri, localName, qName, attributes);
		if (ISMLConstants.COMPLEXTYPE_ELEMENT.equals(localName))
		{
			currentComplexType = super.getPrefix() + attributes.getValue(ISMLConstants.NAME_ATTRIBUTE);
			if (currentComplexType != null && attributes != null)
			{
				Hashtable complexTypeAtts = new Hashtable();
				String acyclic = attributes.getValue(ISMLConstants.ACYCLIC_ATTRIBUTE);
				if (acyclic != null)
					complexTypeAtts.put(ISMLConstants.ACYCLIC_ATTRIBUTE, acyclic);
				
				this.attributes.put(currentComplexType, complexTypeAtts);
			}
		}
		else if ((currentComplexType != null) && (ISMLConstants.EXTENSION_ELEMENT.equals(localName)))
		{
			String base = attributes.getValue(ISMLConstants.BASE_ATTRIBUTE);
			
			/* Before adding the child-parent relation, we need to make sure that the relationship
			 * is valid (e.g. a child type that has acyclic set to false cannot extend a parent 
			 * with acyclic set to true) -- making the assumption that parent types are declared
			 * before child types */
			validateRelationship(base, currentComplexType);
			inheritanceMap.put(currentComplexType, base);
		}
	}
	


	/**
	 * @see org.xml.sax.helpers.DefaultHandler#endElement(java.lang.String, java.lang.String, java.lang.String)
	 */	
	public void endElement(String uri, String localName, String qName) throws SAXException 
	{
		super.endElement(uri, localName, qName);
		if (ISMLConstants.COMPLEXTYPE_ELEMENT.equals(localName))
		{
			currentComplexType = null;
		}		
	}


	/**
	 * Based on the inheritance map passed in, this method will return true if and only if
	 * refTypeInQuestion is a derived type of refType.
	 * 
	 * @param inheritanceMap The inheritance map
	 * @param refType A primary ref type 
	 * @param refTypeInQuestion The reference type in question
	 * @return true iff refTypeInQuestion is a derived type of refType; false otherwise
	 */
	public static boolean isDerivedType(Map inheritanceMap, String refType, String refTypeInQuestion)
	{
		/* A type is always a derived type of itself */
		if (refType.equals(refTypeInQuestion))
			return true;
		
		String currentType = refTypeInQuestion;
		while ((currentType = (String)inheritanceMap.get(currentType)) != null)
		{
			if (refType.equals(currentType))
				return true;
		}
		
		return false;
	}

	/**
	 * @see org.eclipse.cosmos.rm.validation.internal.databuilders.AbstractDataBuilder#getPhase()
	 */
	public byte getPhase() 
	{
		return ISMLConstants.GENICS_PHASE;
	}
	
	/**
	 * Validates the relationship between a parent and a child type
	 * 
	 * @param baseType The parent type
	 * @param childType The child type
	 */
	private void validateRelationship(String baseType, String childType)
	{
		if (isTypeAcyclic(baseType, false) && !isTypeAcyclic(childType, true))
		{
			setStructureValidity(false);
			appendToErrorMessage(ValidationMessages.acyclicBadDerivation);
		}
	}
	
	
	/**
	 * Indicates if a specific type is acyclic
	 * 
	 * @param type The type
	 * @param defaultValue The default value that will be re
	 * @return true if 'type' is acyclic; false otherwise
	 */
	private boolean isTypeAcyclic(String type, boolean defaultValue)
	{
		Map typeAttributes = (Map)attributes.get(type);
		if (typeAttributes != null)
		{
			String acyclicValue = (String)typeAttributes.get(ISMLConstants.ACYCLIC_ATTRIBUTE);
			if (acyclicValue != null)
				return IValidationConstants.TRUE_VALUE.equalsIgnoreCase(acyclicValue);
		}
		
		return defaultValue;
	}
}
