/**********************************************************************
 * Copyright (c) 2006 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * IBM - Initial API and implementation
 **********************************************************************/
package org.eclipse.cosmos.rm.validation.internal.smlvalidators;

import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.cosmos.rm.validation.internal.artifacts.ElementNode;
import org.eclipse.cosmos.rm.validation.internal.common.IValidationOutput;
import org.eclipse.cosmos.rm.validation.internal.core.AbstractSMLValidator;
import org.eclipse.cosmos.rm.validation.internal.databuilders.ComplexTypeElementBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.DataBuilderRegistry;
import org.eclipse.cosmos.rm.validation.internal.databuilders.IDataBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.RefTypeElementBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.RootElementAliasLookupBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.SubstitutionBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.TargetElementBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.TargetElementInstanceBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.TypeInheritanceDataBuilderImpl;
import org.eclipse.cosmos.rm.validation.internal.databuilders.XSElementBuilder;
import org.eclipse.cosmos.rm.validation.internal.util.ParserHelper;

/**
 * This class is used to validate the sml:targetElement, sml:targetRequired and sml:targetType sml extensions. 
 * @author sleeloy
 *
 */
public class TargetValidator extends AbstractSMLValidator
{

	public void initialize(Map validationAttribute)
	{
		super.initialize(validationAttribute);
		DataBuilderRegistry builderRegistry = DataBuilderRegistry.instance();
		
		builderRegistry.registerDataStructureBuilder(TypeInheritanceDataBuilderImpl.ID, new TypeInheritanceDataBuilderImpl());
		builderRegistry.registerDataStructureBuilder(TargetElementBuilder.ID, new TargetElementBuilder());
		builderRegistry.registerDataStructureBuilder(RefTypeElementBuilder.ID, new RefTypeElementBuilder());
		builderRegistry.registerDataStructureBuilder(ComplexTypeElementBuilder.ID, new ComplexTypeElementBuilder());
		builderRegistry.registerDataStructureBuilder(TargetElementInstanceBuilder.ID, new TargetElementInstanceBuilder());
		builderRegistry.registerDataStructureBuilder(RootElementAliasLookupBuilder.ID, new RootElementAliasLookupBuilder());		
		builderRegistry.registerDataStructureBuilder(XSElementBuilder.ID, new XSElementBuilder());
		builderRegistry.registerDataStructureBuilder(SubstitutionBuilder.ID, new SubstitutionBuilder());
						
	}

	public boolean validate()
	{
		IValidationOutput validationLogger = getValidationOutput();
		IDataBuilder targetElement = DataBuilderRegistry.instance().getDataStructureBuilder(TargetElementBuilder.ID);
		IDataBuilder refTypeElement = DataBuilderRegistry.instance().getDataStructureBuilder(RefTypeElementBuilder.ID);
		IDataBuilder typeInheritance = DataBuilderRegistry.instance().getDataStructureBuilder(TypeInheritanceDataBuilderImpl.ID);
		IDataBuilder complexTypeElement = DataBuilderRegistry.instance().getDataStructureBuilder(ComplexTypeElementBuilder.ID);
		IDataBuilder targetInstanceElement  = DataBuilderRegistry.instance().getDataStructureBuilder(TargetElementInstanceBuilder.ID);
		IDataBuilder rootElementAliasElement  = DataBuilderRegistry.instance().getDataStructureBuilder(RootElementAliasLookupBuilder.ID);
		IDataBuilder xsElementElement  = DataBuilderRegistry.instance().getDataStructureBuilder(XSElementBuilder.ID);
		IDataBuilder substitionElement  = DataBuilderRegistry.instance().getDataStructureBuilder(SubstitutionBuilder.ID);
						
		Map targetElems = (Map)targetElement.getDataStructure();
		Map refTypeElems = (Map)refTypeElement.getDataStructure();
		Map complexTypeElems = (Map)complexTypeElement.getDataStructure();
		Map inheritance = (Map)typeInheritance.getDataStructure();
		Map targetInstanceElems = (Map)targetInstanceElement.getDataStructure();
		Map rootElementAliasElems = (Map)rootElementAliasElement.getDataStructure();
		Map xsElementElems = (Map)xsElementElement.getDataStructure();
		Map substitionElementElems = (Map)substitionElement.getDataStructure();
		
		//The sml:targetElement attribute is supported on element 
		//declarations whose type is sml:refType or a derived type
		//of sml:refType. 
		Iterator iter = targetElems.keySet().iterator();
		while (iter.hasNext()){
			Object targetElem = iter.next();
			//check if elem is of type sml:refType
			if (refTypeElems.get(targetElem) == null){
				boolean valid = false;
				//check inheritance tree
				Object parent = inheritance.get(targetElem);
				while (parent != null){
					if (refTypeElems.get(parent) != null){
						valid = true;
						break;
					}
					parent = inheritance.get(parent);
					
				}
				if (!valid){
					validationLogger.reportError(ValidationMessages.bind(ValidationMessages.targetElementNotRefType,targetElem));
					//stop validating
					return false;
				}
			}
			
			//make sure the target value is a global type
			String targetValue = (String)targetElems.get(targetElem);
			targetValue = ParserHelper.removeNameSpace(targetValue);
			
			//get type
			ElementNode xsElement = (ElementNode)xsElementElems.get(targetValue);
			String type = ParserHelper.removeNameSpace(xsElement.getType());
			//check if the type is a global type
			if (complexTypeElems.get(type) == null){
				
				boolean valid = false;
				//check inheritance tree
				String parent = (String)inheritance.get(type);
				parent = ParserHelper.removeNameSpace(parent);
				while (parent != null){
					if (complexTypeElems.get(parent) != null){
						valid = true;
						break;
					}
					parent = (String)inheritance.get(parent);
					parent = ParserHelper.removeNameSpace(parent);
					
				}
				if (!valid){
					validationLogger.reportError(ValidationMessages.bind(ValidationMessages.targetElementNotDefined,targetElem,parent));
					//stop validating
					return false;
				}
				if (!valid){
					validationLogger.reportError(ValidationMessages.bind(ValidationMessages.targetElementNotDefined, targetElem, targetValue));
					//stop validating
					return false;
				}
			}
		}
		
		//make sure
		
		Iterator iter1 = targetInstanceElems.keySet().iterator();
		while (iter1.hasNext()){
			ElementNode key = (ElementNode)iter1.next();
			//get GTE
			Object GTE = ParserHelper.removeNameSpace((String)targetElems.get(key.getName()));
			//get target ref.  Note:This should not be null
			Object targetRef = targetInstanceElems.get(key);
			//check to see if the target ref is instance of GTE
			if (targetRef.equals(GTE)){
				continue;
			}
			else{
				//the ref could be an alias
				//Search through alias datastructure for actual element
				Iterator iter2 = rootElementAliasElems.keySet().iterator();
				while (iter2.hasNext()){
					List aliasList = (List)iter2.next();
					if (aliasList.contains(targetRef)){
						targetRef = rootElementAliasElems.get(aliasList);
						break;
					}
				}
				if (targetRef.equals(GTE)){
					continue;
				}
				//there could be a substition group
				Object substitionGroup = substitionElementElems.get(targetRef);				
				boolean valid = false;
				while (substitionGroup != null){
					substitionGroup = ParserHelper.removeNameSpace((String)substitionGroup);
					if (substitionGroup.equals(GTE)){
						valid = true;
						break;
					}			
					substitionGroup = substitionElementElems.get(substitionGroup);
				}
				if (valid) continue;
			}			
			validationLogger.reportError(ValidationMessages.bind(ValidationMessages.invalidTargetInstance, targetRef, GTE));		 //$NON-NLS-2$
			return false;
		}
		
			
		return true;
	}

}
